/*
 * _____________________________________________________________________________
 * 
 * INAF - OATS National Institute for Astrophysics - Astronomical Observatory of
 * Trieste INAF - IA2 Italian Center for Astronomical Archives
 * _____________________________________________________________________________
 * 
 * Copyright (C) 2017 Istituto Nazionale di Astrofisica
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License Version 3 as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
package it.inaf.ia2.tsm.datalayer;

import it.inaf.ia2.tsm.ColumnHolder;
import it.inaf.ia2.tsm.ConsistencyChecks;
import it.inaf.ia2.tsm.Key;
import it.inaf.ia2.tsm.TapSchema;
import it.inaf.ia2.tsm.TapSchemaEntity;
import it.inaf.ia2.tsm.model.ColumnModel;
import it.inaf.ia2.tsm.model.TableModel;
import it.inaf.ia2.tsm.model.SchemaModel;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This interface can be seen as a DAO (Data Access Object pattern), however it
 * has been called more generically a "broker to the database", because it
 * doesn't map exactly the CRUD operations, indeed it contains some specific
 * methods for interacting to the databases (executing SQL statement) and
 * retrieving/storing data regarding TAP SCHEMA. Generic operations are
 * implemented inside the abstract class {@link DBBrokerTemplate}. The latter is
 * then extended by specific MySQL and Postgres Implementations.
 *
 * @author Sonia Zorba {@literal <zorba at oats.inaf.it>}
 */
public interface DBBroker {

    /**
     * Returns the name of all schemata contained into the source database.
     */
    List<String> getAllSchemaNames() throws SQLException;

    /**
     * Returns the name of all schemata which structure is compatible with a
     * TAP_SCHEMA schema structure.
     */
    List<String> getAllTAPSchemaNames(List<String> allSchemas) throws SQLException;

    /**
     * Try to guess the version of an existing TAP_SCHEMA analyzing its shallow
     * structure (e.<!-- -->g.<!-- --> the presence of the set of columns
     * defined into the related XML model).
     *
     * @return the name of the guessed TAP_SCHEMA version.
     */
    String detectVersion(String tapSchemaName) throws SQLException;

    /**
     * Returns a list of all the schemata exposed by an existing TAP_SCHEMA.
     * Queries the {@code TAP_SCHEMA.schemas} schema.
     */
    List<String> getExposedSchemas(String tapSchemaName) throws SQLException;

    /**
     * Returns all the tables owned by a schema, given its name.
     */
    List<String> getAllTablesNames(String schemaName) throws SQLException;

    /**
     * Returns the table type ("table" or "view") for all the tables owned by a
     * schema, given its name.
     *
     * @return a {@code Map} having table names as keys and table types as
     * values.
     */
    Map<String, String> getAllTableTypes(String schemaName) throws SQLException;

    /**
     * Returns the list of the names of the columns contained in a given table.
     */
    List<String> getAllColumnsNames(String schemaName, String tableName) throws SQLException;

    /**
     * Returns the metadata of all the columns contained in a given table.
     *
     * @param schemaName the name of the schema.
     * @param tableSimpleName the name of the table.
     * @param tableModel the {@link TableModel} of the owner column if it is a
     * TAP_SCHEMA table or the ObsCore table, null otherwise. This is used for
     * consistency checking on these tables datatype.
     * @param dataTypeMode the {@link DataTypeMode} used in the selected
     * TAP_SCHEMA version; this is used only if the previous parameter is not
     * null.
     * @return a {@code Map} having the column names as keys and the column
     * metadata as values.
     */
    Map<String, Map<String, Object>> getAllColumnsMetadata(String schemaName, String tableSimpleName, TableModel tableModel, DataTypeMode dataTypeMode) throws SQLException;

    List<Key> getKeys(TapSchema tapSchema, String schemaName, String realSchemaName) throws SQLException;

    List<Map<String, Object>> getSavedItems(String tapSchemaName, TableModel tableModel, String whereCondition, Object[] whereParams) throws SQLException;

    List<Map<String, Object>> getSavedItems(String tapSchemaName, TableModel tableModel) throws SQLException;

    void insertItem(String tapSchemaName, TapSchemaEntity entity, Connection conn) throws SQLException;

    void updateItem(String tapSchemaName, TapSchemaEntity entity, Connection conn, String whereCondition, Object... whereParams) throws SQLException;

    void createTapSchemaStructure(String tapSchemaName, SchemaModel tapSchemaModel) throws SQLException;

    void createIvoaSchemaStructure(SchemaModel ivoaSchemaModel, String realIvoaSchemaName) throws SQLException;

    void save(TapSchema tapSchema) throws SQLException;

    void createTable(String schemaName, TableModel tableModel) throws SQLException;

    void addColumn(ColumnHolder columnHolder, ColumnModel columnModel) throws SQLException;

    /**
     * @param like is for using both the schema name and the complete table name
     * in the query
     */
    Set<String> getKeysToRemove(String tapSchemaName, String like) throws SQLException;

    void deleteUnexistingEntities(String tapSchemaName, ConsistencyChecks consistencyChecks, Set<String> keysToRemoveIds) throws SQLException;

    Set<String> getKeysToRemoveFromUnexistingColumn(String tapSchemaName, ColumnHolder unexistingColumn) throws SQLException;

    void updateTapSchemaColumnValue(String tapSchemaName, String completeTableName, String columnName, String key, Object value) throws SQLException;

    void alterDataType(String schemaName, String tableName, String columnName, String adqlDataType, Integer size) throws SQLException;
}
